(*<*)
theory VDMComplete = VDMSound:
(*>*)

subsection{*Completeness*}

text{*Strongest specifications, given precisely by the operational
behaviour.*}

constdefs SSpec::"IMP \<Rightarrow> VDMAssn"
"SSpec C s t \<equiv> s,C \<Down> t"

text{*Sttrongest specifications are valid \ldots*}
lemma SSpec_valid: "\<Turnstile> C : (SSpec C)"
(*<*)by (simp add: vdm_valid_def SSpec_def)(*>*)

text{*\ldots and imply any other valid assertions for the same program
(hence their name).*}

lemma SSpec_strong: "\<Turnstile> C :A \<Longrightarrow> \<forall> s t . SSpec C s t \<longrightarrow> A s t"
(*<*)by (simp add: vdm_valid_def SSpec_def)(*>*)

text{*By induction on $C$ we show the following.*}

lemma SSpec_derivable:"G \<rhd> Call : SSpec Call \<Longrightarrow> G \<rhd> C : SSpec C"
(*<*)
apply (induct C)
  apply (rule VDMConseq)
  apply (rule VDMSkip) apply (simp add: SSpec_def Sem_def, rule, rule) apply (rule SemSkip)
  apply (rule VDMConseq)
  apply (rule VDMAssign) apply (simp add: SSpec_def Sem_def, rule, rule) apply (rule SemAssign) apply simp
apply clarsimp
  apply (rule VDMConseq)
  apply (rule VDMComp) apply assumption+ apply (simp add: SSpec_def Sem_def, clarsimp) 
  apply rule apply (rule SemComp) apply assumption+ apply simp
apply clarsimp
apply (rule VDMConseq)
  apply (erule VDMWhile) 
    prefer 3 apply (subgoal_tac "\<forall>s t. SSpec (While BExpr C) s t \<and> \<not> evalB BExpr t \<longrightarrow> SSpec (While BExpr C) s t", assumption)
    apply simp
    apply (simp add: SSpec_def Sem_def, clarsimp) apply (rule, erule SemWhileF) apply simp
    apply (simp add: SSpec_def Sem_def, clarsimp) apply (rule, erule SemWhileT) apply assumption+ apply simp
apply clarsimp
apply (rule VDMConseq)
  apply (rule VDMIff) apply assumption+ apply (simp add: SSpec_def Sem_def, clarsimp)
  apply (erule thin_rl, erule thin_rl) 
  apply (case_tac "evalB BExpr s")
  apply clarsimp apply (rule, rule SemTrue) apply assumption+ 
  apply clarsimp apply (rule, rule SemFalse) apply (assumption, assumption)
apply assumption 
done
(*>*)

text{*The (singleton) strong context contains the strongest
specification for the procedure.*}

constdefs StrongG :: "VDMAssn set"
  "StrongG \<equiv> {SSpec Call}"

text{*By constructions, the strongest specification of the procedure's
body can be verified with respect to this context.*}

lemma StrongG_Body: "StrongG \<rhd> body : SSpec Call"
(*<*)
apply (subgoal_tac "StrongG \<rhd> body : SSpec body")
  apply (erule VDMConseq) apply (simp add: SSpec_def Sem_def, clarsimp)
  apply (rule, erule SemCall)
apply (rule SSpec_derivable) apply (rule VDMAx) apply (simp add: StrongG_def)
done
(*>*)

text{*Thus, the strong context is verified.*}

lemma StrongG_verified: "verified StrongG"
(*<*)
apply (simp add: verified_def)
apply (rule allI)+
apply rule
apply (subgoal_tac "StrongG \<rhd> body : SSpec Call")
apply (simp add: StrongG_def)
apply (rule StrongG_Body)
done
(*>*)

text{*Using this result and the rules \emph{cut} and \emph{Mutrec}, we
show that arbitrary commands satisfy their strongest specification
with respect to the empty context.*}

lemma SSpec_derivable_empty:"{} \<rhd> C : SSpec C"
(*<*)
apply (rule cut)
apply (rule MUTREC) apply (subgoal_tac "finite StrongG", assumption)
  apply (simp add: StrongG_def)
  apply (simp add: StrongG_def)
  apply (rule StrongG_verified)
  apply (simp add: StrongG_def)
  apply (rule SSpec_derivable) apply (rule VDMAx) apply simp
done
(*>*)

text{*From this, we easily obtain (relative) completeness.*}

theorem VDM_Complete: "\<Turnstile> C : A \<Longrightarrow> {} \<rhd> C : A"
(*<*)
apply (rule VDMConseq)
apply (rule SSpec_derivable_empty)
apply (erule SSpec_strong)
done
(*>*)

text{*Furthermore, valid contexts are verified*}

lemma Ctxt_valid_verified: "vdmctxt_valid G \<Longrightarrow> verified G"
(*<*)
apply (simp add: vdmctxt_valid_def verified_def, clarsimp)
apply (erule_tac x=A in allE, simp)
apply (subgoal_tac "\<Turnstile> body : A")
apply (rotate_tac 1, erule thin_rl, drule VDM_Complete) apply (erule VDM_CtxtSub) apply fast
apply (simp add: vdm_valid_def Sem_def, clarsimp)
apply (erule_tac x=s in allE, erule_tac x=t in allE, erule mp)
apply (rule, erule SemCall)
done
(*>*)

(*<*)
end
(*>*)
